import jax.numpy as jnp
# from jax import jit, vmap, tree_util
# from functools import partial # for JAX jit with static params

from math import floor, ceil
from .chiphifunc import *

# Sum: implemented as a function taking in a single-argument func and the lower/upper bounds
# expr should be non-dynamic.
# Non-jitted because an argument is a callable. A wrapper for a py_sum with no
# callable argument can be jitted.
def py_sum(expr, lower:int, upper:int):
    # The integer 0 cannot be added to even ChiPhiFuncs,
    # because JAX does not support conditionals on traced arguments.
    out = ChiPhiFuncSpecial(0)
    upper_floor = floor(upper)
    lower_ceil = ceil(lower)
    # If lower==upper then return expr(lower)
    if upper_floor==lower_ceil:
        return(expr(lower_ceil))
    # Warning for lower>upper
    if lower_ceil>upper_floor:
        # This is classified as "out of bound".
        # Originally the code is -1. Since the formula
        # are checked correct, these are made 0.
        return(ChiPhiFuncSpecial(0))
    indices = list(range(lower_ceil,upper_floor+1))
    out_list = jax.tree_util.tree_map(expr, indices)
    for item in out_list:
        out = out+item
    return(out)

# In the JAX implementation, there is no distinction between how the outmost and
# inner sums are evaluated.
py_sum_parallel = py_sum

## Condition operators

# Used to make sure new indices of terms and new upper bounds are within the
# bound of the original summations
# is_seq(a,b): 1 if a<=b
# @partial(jit, static_argnums=(0, 1,))
def is_seq(a, b):
    if a<=b:
        return(1)
    else:
        return(ChiPhiFuncSpecial(0))
# Used to ensure new index values (after removing the innermost sum) are integers.
# is_integer(a): 1 if a is integer
# @partial(jit, static_argnums=(0,))
def is_integer(a):
    if a%1==0:
        return(1)
    else:
        return(ChiPhiFuncSpecial(0))

# @partial(jit, static_argnums=(1, 2, ))
def diff_backend(y, is_chi:bool, order):
    '''
    Takes phi or chi derivative.
    Input:
    y: ChiPhiFunc or const
    is_chi: True for 'chi' or False for 'phi'
    order: order of derivative
    '''
    if jnp.isscalar(y):
        return(0)
    out = y
    if isinstance(y, ChiPhiFunc):
        if is_chi:
            out = out.dchi(order)
        else:
            out = out.dphi(order)
    else:
        return(ChiPhiFuncSpecial(-13))
    return(out)

# Maxima sometimes merges a few diff's together.
# @partial(jit, static_argnums=(1, 2, 3, 4,))
def diff(y, is_chi1:bool, order1:int, is_chi2=None, order2=None):
    out = diff_backend(y, is_chi1, order1)
    if is_chi2 is not None:
        out = diff_backend(out, is_chi2, order2)
    return(out)

# Faster. In this case, tensordot is faster than einsum.
def einsum_ijkl_jmln_to_imkn(array_A, array_B):
    if len(array_A.shape)!=4 or len(array_B.shape)!=4:
        return(jnp.nan)# Both input need to be 4d arrays
    # ikjl
    A_transposed = jnp.transpose(array_A, (0,2,1,3))
    # jlmn
    B_transposed = jnp.transpose(array_B, (0,2,1,3))
    # ikmn
    array_out = jnp.tensordot(A_transposed, B_transposed)
    return(jnp.transpose(array_out, (0,2,1,3)))



from aqsc.math_utilities import *
from aqsc.chiphifunc import *
# from jax import jit
# from functools import partial
# @partial(jit, static_argnums=(0, 10, 11, 12, 13))
def eval_B_psi_coefs_D3(n_eval, X_coef_cp, Y_coef_cp,
    Delta_coef_cp, B_alpha_coef, B_denom_coef_c,
    dl_p, tau_p, kap_p, iota_coef,
    num_mode, cap_axis0, len_tensor, nfp):

    coef_B_psi_dphi_0_dchi_0_all_but_Y_D3 = (((diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1)-Y_coef_cp[1]*diff(X_coef_cp[1],True,2))*diff(Y_coef_cp[1],False,1)+iota_coef[0]*X_coef_cp[1]*Y_coef_cp[1]*diff(Y_coef_cp[1],True,3)+X_coef_cp[1]*Y_coef_cp[1]*diff(Y_coef_cp[1],True,2,False,1)+((-iota_coef[0]*X_coef_cp[1]*diff(Y_coef_cp[1],True,1))+Y_coef_cp[1]*diff(X_coef_cp[1],False,1)+iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,2)-X_coef_cp[1]*diff(Y_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1,False,1)-diff(X_coef_cp[1],False,1)*(diff(Y_coef_cp[1],True,1))**2+Y_coef_cp[1]*diff(X_coef_cp[1],True,1,False,1)*diff(Y_coef_cp[1],True,1)-iota_coef[0]*Y_coef_cp[1]**2*diff(X_coef_cp[1],True,3)-Y_coef_cp[1]**2*diff(X_coef_cp[1],True,2,False,1))*dl_p*kap_p*tau_p+((diff(X_coef_cp[1],True,1))**2*diff(Y_coef_cp[1],False,1)-iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,2)-X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1,False,1)-diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1)*diff(Y_coef_cp[1],True,1)+iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,2)+Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,1,False,1))*diff(kap_p,False,1)+((-X_coef_cp[1]**2*diff(Y_coef_cp[1],True,2))+X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1)+X_coef_cp[1]*Y_coef_cp[1]*diff(X_coef_cp[1],True,2)-Y_coef_cp[1]*(diff(X_coef_cp[1],True,1))**2)*dl_p**2*kap_p**3+((-(diff(X_coef_cp[1],True,1))**2*diff(Y_coef_cp[1],False,2))+((-diff(X_coef_cp[1],True,2)*diff(X_coef_cp[1],False,1))-3*iota_coef[0]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,2)-2*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,1,False,1))*diff(Y_coef_cp[1],False,1)+(iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],False,1)+2*iota_coef[0]**2*X_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,3)+(X_coef_cp[1]*diff(X_coef_cp[1],False,1)+3*iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,2,False,1)+((diff(X_coef_cp[1],False,1))**2+4*iota_coef[0]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1)+2*iota_coef[0]**2*(diff(X_coef_cp[1],True,1))**2)*diff(Y_coef_cp[1],True,2)+X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1,False,2)+2*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1)*diff(Y_coef_cp[1],True,1,False,1)+(diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,2)-iota_coef[0]*diff(X_coef_cp[1],True,2)*diff(X_coef_cp[1],False,1)-2*iota_coef[0]**2*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,2))*diff(Y_coef_cp[1],True,1)+((-iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,3))-Y_coef_cp[1]*diff(X_coef_cp[1],True,2,False,1))*diff(X_coef_cp[1],False,1)-2*iota_coef[0]**2*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,3)-3*iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,2,False,1)-Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,1,False,2))*kap_p)/(dl_p*kap_p**2*n_eval-dl_p*kap_p**2)

    coef_B_psi_dphi_0_dchi_1_all_but_Y_D3 = -((Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],False,1)-2*iota_coef[0]*X_coef_cp[1]*Y_coef_cp[1]*diff(Y_coef_cp[1],True,2)-X_coef_cp[1]*Y_coef_cp[1]*diff(Y_coef_cp[1],True,1,False,1)+iota_coef[0]*X_coef_cp[1]*(diff(Y_coef_cp[1],True,1))**2+((-Y_coef_cp[1]*diff(X_coef_cp[1],False,1))-iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,1)+2*iota_coef[0]*Y_coef_cp[1]**2*diff(X_coef_cp[1],True,2)+Y_coef_cp[1]**2*diff(X_coef_cp[1],True,1,False,1))*dl_p*kap_p*tau_p+(iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1)-iota_coef[0]*Y_coef_cp[1]*(diff(X_coef_cp[1],True,1))**2)*diff(kap_p,False,1)+(X_coef_cp[1]**2*diff(Y_coef_cp[1],True,1)-X_coef_cp[1]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1))*dl_p**2*kap_p**3+((diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1)+3*iota_coef[0]*(diff(X_coef_cp[1],True,1))**2)*diff(Y_coef_cp[1],False,1)+((-2*iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],False,1))-4*iota_coef[0]**2*X_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,2)+((-X_coef_cp[1]*diff(X_coef_cp[1],False,1))-3*iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,1,False,1)+((-(diff(X_coef_cp[1],False,1))**2)-3*iota_coef[0]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1))*diff(Y_coef_cp[1],True,1)+(2*iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,2)+Y_coef_cp[1]*diff(X_coef_cp[1],True,1,False,1))*diff(X_coef_cp[1],False,1)+4*iota_coef[0]**2*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,2)+3*iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,1,False,1))*kap_p)/(dl_p*kap_p**2*n_eval-dl_p*kap_p**2)

    coef_B_psi_dphi_0_dchi_2_all_but_Y_D3 = ((iota_coef[0]*X_coef_cp[1]*Y_coef_cp[1]*diff(Y_coef_cp[1],True,1)-iota_coef[0]*Y_coef_cp[1]**2*diff(X_coef_cp[1],True,1))*dl_p*tau_p+(iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],False,1)+2*iota_coef[0]**2*X_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,1)-iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1)-2*iota_coef[0]**2*Y_coef_cp[1]*(diff(X_coef_cp[1],True,1))**2)/(dl_p*kap_p*n_eval-dl_p*kap_p)

    coef_B_psi_dphi_0_dchi_3_all_but_Y_D3 = ChiPhiFuncSpecial(0)

    coef_B_psi_dphi_1_dchi_0_all_but_Y_D3 = ((X_coef_cp[1]*Y_coef_cp[1]*diff(Y_coef_cp[1],True,2)-X_coef_cp[1]*(diff(Y_coef_cp[1],True,1))**2+Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1)-Y_coef_cp[1]**2*diff(X_coef_cp[1],True,2))*dl_p*kap_p*tau_p+(Y_coef_cp[1]*(diff(X_coef_cp[1],True,1))**2-X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1))*diff(kap_p,False,1)+((-2*(diff(X_coef_cp[1],True,1))**2*diff(Y_coef_cp[1],False,1))+(X_coef_cp[1]*diff(X_coef_cp[1],False,1)+3*iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,2)+2*X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1,False,1)+2*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1)*diff(Y_coef_cp[1],True,1)-Y_coef_cp[1]*diff(X_coef_cp[1],True,2)*diff(X_coef_cp[1],False,1)-3*iota_coef[0]*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,2)-2*Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],True,1,False,1))*kap_p)/(dl_p*kap_p**2*n_eval-dl_p*kap_p**2)

    coef_B_psi_dphi_1_dchi_1_all_but_Y_D3 = ((X_coef_cp[1]*Y_coef_cp[1]*diff(Y_coef_cp[1],True,1)-Y_coef_cp[1]**2*diff(X_coef_cp[1],True,1))*dl_p*tau_p+(X_coef_cp[1]*diff(X_coef_cp[1],False,1)+3*iota_coef[0]*X_coef_cp[1]*diff(X_coef_cp[1],True,1))*diff(Y_coef_cp[1],True,1)-Y_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(X_coef_cp[1],False,1)-3*iota_coef[0]*Y_coef_cp[1]*(diff(X_coef_cp[1],True,1))**2)/(dl_p*kap_p*n_eval-dl_p*kap_p)

    coef_B_psi_dphi_1_dchi_2_all_but_Y_D3 = ChiPhiFuncSpecial(0)

    coef_B_psi_dphi_1_dchi_3_all_but_Y_D3 = ChiPhiFuncSpecial(0)

    coef_B_psi_dphi_2_dchi_0_all_but_Y_D3 = (X_coef_cp[1]*diff(X_coef_cp[1],True,1)*diff(Y_coef_cp[1],True,1)-Y_coef_cp[1]*(diff(X_coef_cp[1],True,1))**2)/(dl_p*kap_p*n_eval-dl_p*kap_p)

    coef_B_psi_dphi_2_dchi_1_all_but_Y_D3 = ChiPhiFuncSpecial(0)

    coef_B_psi_dphi_2_dchi_2_all_but_Y_D3 = ChiPhiFuncSpecial(0)

    coef_B_psi_dphi_2_dchi_3_all_but_Y_D3 = ChiPhiFuncSpecial(0)
    return(
        to_tensor_fft_op_multi_dim(coef_B_psi_dphi_0_dchi_0_all_but_Y_D3,
            dphi=0, dchi=0,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_0_dchi_1_all_but_Y_D3,
            dphi=0, dchi=1,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_0_dchi_2_all_but_Y_D3,
            dphi=0, dchi=2,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_0_dchi_3_all_but_Y_D3,
            dphi=0, dchi=3,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_1_dchi_0_all_but_Y_D3,
            dphi=1, dchi=0,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_1_dchi_1_all_but_Y_D3,
            dphi=1, dchi=1,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_1_dchi_2_all_but_Y_D3,
            dphi=1, dchi=2,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_1_dchi_3_all_but_Y_D3,
            dphi=1, dchi=3,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_2_dchi_0_all_but_Y_D3,
            dphi=2, dchi=0,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_2_dchi_1_all_but_Y_D3,
            dphi=2, dchi=1,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_2_dchi_2_all_but_Y_D3,
            dphi=2, dchi=2,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +to_tensor_fft_op_multi_dim(coef_B_psi_dphi_2_dchi_3_all_but_Y_D3,
            dphi=2, dchi=3,
            num_mode=num_mode, cap_axis0=cap_axis0,
            len_tensor=len_tensor, nfp=nfp)
        +0
    )



# Evaluates Zn+1. Requires X[..., n], Y[..., n], Z[..., n],
# B_theta_coef_cp[..., n], B_psi_coef_cp[..., n-1],
# B_alpha_coef [..., (n-1)/2 or (n-2)/2]
# iota_coef [..., (n-1)/2 or (n-2)/2]
# kap_p, dl_p, tau_p
from math import floor, ceil
from aqsc.math_utilities import *
def eval_Znp1_cp(n, X_coef_cp, Y_coef_cp, Z_coef_cp, \
    B_theta_coef_cp, B_psi_coef_cp, \
    B_alpha_coef, \
    kap_p, dl_p, tau_p, iota_coef):
    def sum_arg_62(i162):
        # Child args for sum_arg_62
        def sum_arg_61(i160):
            # Child args for sum_arg_61
            return(B_theta_coef_cp[i160]*Y_coef_cp[n-i162-i160+2])

        return(i162*Z_coef_cp[i162]*py_sum(sum_arg_61,0,n-i162+2))

    def sum_arg_60(i158):
        # Child args for sum_arg_60
        def sum_arg_59(i156):
            # Child args for sum_arg_59
            return(B_psi_coef_cp[i156]*Y_coef_cp[n-i158-i156])

        return(diff(Z_coef_cp[i158],True,1)*py_sum(sum_arg_59,0,n-i158))

    def sum_arg_58(i122):
        # Child args for sum_arg_58
        def sum_arg_57(i120):
            # Child args for sum_arg_57
            return(B_theta_coef_cp[i120]*X_coef_cp[n-i122-i120+2])

        return(i122*Z_coef_cp[i122]*py_sum(sum_arg_57,0,n-i122+2))

    def sum_arg_56(i118):
        # Child args for sum_arg_56
        def sum_arg_55(i116):
            # Child args for sum_arg_55
            return(B_psi_coef_cp[i116]*X_coef_cp[n-i118-i116])

        return(diff(Z_coef_cp[i118],True,1)*py_sum(sum_arg_55,0,n-i118))

    def sum_arg_54(i736):
        # Child args for sum_arg_54
        def sum_arg_53(i184):
            # Child args for sum_arg_53
            return(B_psi_coef_cp[i184]*diff(X_coef_cp[n-i736-i184],True,1))

        return(diff(Z_coef_cp[i736],False,1)*py_sum(sum_arg_53,0,n-i736))

    def sum_arg_52(i730):
        # Child args for sum_arg_52
        def sum_arg_51(i188):
            # Child args for sum_arg_51
            return(B_psi_coef_cp[i188]*diff(X_coef_cp[n-i730-i188],False,1))

        return(diff(Z_coef_cp[i730],True,1)*py_sum(sum_arg_51,0,n-i730))

    def sum_arg_50(i262):
        # Child args for sum_arg_50
        def sum_arg_49(i140):
            # Child args for sum_arg_49
            return(B_psi_coef_cp[i140]*diff(Y_coef_cp[n-i262-i140],False,1))

        return(diff(Z_coef_cp[i262],True,1)*py_sum(sum_arg_49,0,n-i262))

    def sum_arg_48(i256):
        # Child args for sum_arg_48
        def sum_arg_47(i136):
            # Child args for sum_arg_47
            return(B_psi_coef_cp[i136]*diff(Y_coef_cp[n-i256-i136],True,1))

        return(diff(Z_coef_cp[i256],False,1)*py_sum(sum_arg_47,0,n-i256))

    def sum_arg_46(i733):
        # Child args for sum_arg_46
        def sum_arg_45(i734):
            # Child args for sum_arg_45
            return((((-i734)+2*i733-2)*diff(Z_coef_cp[i734],True,1)*X_coef_cp[(-n)-i734+2*i733-2]-diff(Z_coef_cp[i734],True,1)*X_coef_cp[(-n)-i734+2*i733-2]*n)*is_seq(n-i733+2,i733-i734))

        return(is_seq(0,n-i733+2)*B_alpha_coef[n-i733+2]*is_integer(n-i733+2)*py_sum(sum_arg_45,0,i733))

    def sum_arg_44(i732):
        # Child args for sum_arg_44
        return(B_psi_coef_cp[i732]*diff(X_coef_cp[n-i732],True,1))

    def sum_arg_43(i726):
        # Child args for sum_arg_43
        def sum_arg_42(i186):
            # Child args for sum_arg_42
            return((B_theta_coef_cp[i186]*n-B_theta_coef_cp[i186]*i726+(2-i186)*B_theta_coef_cp[i186])*X_coef_cp[n-i726-i186+2])

        return(diff(Z_coef_cp[i726],False,1)*py_sum(sum_arg_42,0,n-i726+2))

    def sum_arg_41(i723):
        # Child args for sum_arg_41
        def sum_arg_40(i724):
            # Child args for sum_arg_40
            return((((-i724)+2*i723-2)*diff(X_coef_cp[i724],True,1)*Z_coef_cp[(-n)-i724+2*i723-2]-diff(X_coef_cp[i724],True,1)*Z_coef_cp[(-n)-i724+2*i723-2]*n)*is_seq(n-i723+2,i723-i724))

        return(is_seq(0,n-i723+2)*B_alpha_coef[n-i723+2]*is_integer(n-i723+2)*py_sum(sum_arg_40,0,i723))

    def sum_arg_39(i264):
        # Child args for sum_arg_39
        def sum_arg_38(i138):
            # Child args for sum_arg_38
            return((B_theta_coef_cp[i138]*n-B_theta_coef_cp[i138]*i264+(2-i138)*B_theta_coef_cp[i138])*Y_coef_cp[n-i264-i138+2])

        return(diff(Z_coef_cp[i264],False,1)*py_sum(sum_arg_38,0,n-i264+2))

    def sum_arg_37(i259):
        # Child args for sum_arg_37
        def sum_arg_36(i260):
            # Child args for sum_arg_36
            return((((-i260)+2*i259-2)*diff(Y_coef_cp[i260],True,1)*Z_coef_cp[(-n)-i260+2*i259-2]-diff(Y_coef_cp[i260],True,1)*Z_coef_cp[(-n)-i260+2*i259-2]*n)*is_seq(n-i259+2,i259-i260))

        return(is_seq(0,n-i259+2)*B_alpha_coef[n-i259+2]*is_integer(n-i259+2)*py_sum(sum_arg_36,0,i259))

    def sum_arg_35(i258):
        # Child args for sum_arg_35
        return(B_psi_coef_cp[i258]*diff(Y_coef_cp[n-i258],True,1))

    def sum_arg_34(i251):
        # Child args for sum_arg_34
        def sum_arg_33(i252):
            # Child args for sum_arg_33
            return((((-i252)+2*i251-2)*diff(Z_coef_cp[i252],True,1)*Y_coef_cp[(-n)-i252+2*i251-2]-diff(Z_coef_cp[i252],True,1)*Y_coef_cp[(-n)-i252+2*i251-2]*n)*is_seq(n-i251+2,i251-i252))

        return(is_seq(0,n-i251+2)*B_alpha_coef[n-i251+2]*is_integer(n-i251+2)*py_sum(sum_arg_33,0,i251))

    def sum_arg_32(i194):
        # Child args for sum_arg_32
        def sum_arg_31(i192):
            # Child args for sum_arg_31
            return((B_theta_coef_cp[i192]*n-B_theta_coef_cp[i192]*i194+(2-i192)*B_theta_coef_cp[i192])*Z_coef_cp[n-i194-i192+2])

        return(diff(X_coef_cp[i194],False,1)*py_sum(sum_arg_31,0,n-i194+2))

    def sum_arg_30(i178):
        # Child args for sum_arg_30
        def sum_arg_29(i176):
            # Child args for sum_arg_29
            return(B_theta_coef_cp[i176]*X_coef_cp[n-i178-i176+2])

        return(i178*X_coef_cp[i178]*py_sum(sum_arg_29,0,n-i178+2))

    def sum_arg_28(i174):
        # Child args for sum_arg_28
        def sum_arg_27(i172):
            # Child args for sum_arg_27
            return(B_theta_coef_cp[i172]*Z_coef_cp[n-i174-i172+2])

        return(i174*Z_coef_cp[i174]*py_sum(sum_arg_27,0,n-i174+2))

    def sum_arg_26(i170):
        # Child args for sum_arg_26
        def sum_arg_25(i168):
            # Child args for sum_arg_25
            return(B_psi_coef_cp[i168]*X_coef_cp[n-i170-i168])

        return(diff(X_coef_cp[i170],True,1)*py_sum(sum_arg_25,0,n-i170))

    def sum_arg_24(i166):
        # Child args for sum_arg_24
        def sum_arg_23(i164):
            # Child args for sum_arg_23
            return(B_psi_coef_cp[i164]*Z_coef_cp[n-i166-i164])

        return(diff(Z_coef_cp[i166],True,1)*py_sum(sum_arg_23,0,n-i166))

    def sum_arg_22(i146):
        # Child args for sum_arg_22
        def sum_arg_21(i144):
            # Child args for sum_arg_21
            return((B_theta_coef_cp[i144]*n-B_theta_coef_cp[i144]*i146+(2-i144)*B_theta_coef_cp[i144])*Z_coef_cp[n-i146-i144+2])

        return(diff(Y_coef_cp[i146],False,1)*py_sum(sum_arg_21,0,n-i146+2))

    def sum_arg_20(i130):
        # Child args for sum_arg_20
        def sum_arg_19(i128):
            # Child args for sum_arg_19
            return(B_theta_coef_cp[i128]*X_coef_cp[n-i130-i128+2])

        return(i130*Y_coef_cp[i130]*py_sum(sum_arg_19,0,n-i130+2))

    def sum_arg_18(i126):
        # Child args for sum_arg_18
        def sum_arg_17(i124):
            # Child args for sum_arg_17
            return(B_psi_coef_cp[i124]*X_coef_cp[n-i126-i124])

        return(diff(Y_coef_cp[i126],True,1)*py_sum(sum_arg_17,0,n-i126))

    def sum_arg_16(i737):
        # Child args for sum_arg_16
        def sum_arg_15(i738):
            # Child args for sum_arg_15
            def sum_arg_14(i1193):
                # Child args for sum_arg_14
                return(i1193*Z_coef_cp[i1193]*diff(X_coef_cp[(-i738)+i737-i1193],True,1))

            return(is_seq(0,(-n)+i738+i737-2)*B_theta_coef_cp[(-n)+i738+i737-2]*is_integer((-n)+i738+i737-2)*is_seq((-n)+i738+i737-2,i738)*py_sum(sum_arg_14,0,i737-i738))

        return(iota_coef[n-i737+2]*py_sum(sum_arg_15,0,i737))

    def sum_arg_13(i727):
        # Child args for sum_arg_13
        def sum_arg_12(i728):
            # Child args for sum_arg_12
            def sum_arg_11(i1177):
                # Child args for sum_arg_11
                return(i1177*X_coef_cp[i1177]*diff(Z_coef_cp[(-i728)+i727-i1177],True,1))

            return(is_seq(0,(-n)+i728+i727-2)*B_theta_coef_cp[(-n)+i728+i727-2]*is_integer((-n)+i728+i727-2)*is_seq((-n)+i728+i727-2,i728)*py_sum(sum_arg_11,0,i727-i728))

        return(iota_coef[n-i727+2]*py_sum(sum_arg_12,0,i727))

    def sum_arg_10(i265):
        # Child args for sum_arg_10
        def sum_arg_9(i266):
            # Child args for sum_arg_9
            def sum_arg_8(i722):
                # Child args for sum_arg_8
                return(diff(Z_coef_cp[(-i722)-i266+i265],True,1)*i722*Y_coef_cp[i722])

            return(is_seq(0,(-n)+i266+i265-2)*B_theta_coef_cp[(-n)+i266+i265-2]*is_integer((-n)+i266+i265-2)*is_seq((-n)+i266+i265-2,i266)*py_sum(sum_arg_8,0,i265-i266))

        return(iota_coef[n-i265+2]*py_sum(sum_arg_9,0,i265))

    def sum_arg_7(i253):
        # Child args for sum_arg_7
        def sum_arg_6(i254):
            # Child args for sum_arg_6
            def sum_arg_5(i706):
                # Child args for sum_arg_5
                return(diff(Y_coef_cp[(-i706)-i254+i253],True,1)*i706*Z_coef_cp[i706])

            return(is_seq(0,(-n)+i254+i253-2)*B_theta_coef_cp[(-n)+i254+i253-2]*is_integer((-n)+i254+i253-2)*is_seq((-n)+i254+i253-2,i254)*py_sum(sum_arg_5,0,i253-i254))

        return(iota_coef[n-i253+2]*py_sum(sum_arg_6,0,i253))

    def sum_arg_4(i199):
        # Child args for sum_arg_4
        return(is_seq(0,n-i199)*diff(Y_coef_cp[2*i199-n],True,1)*iota_coef[n-i199]*is_integer(n-i199)*is_seq(n-i199,i199))

    def sum_arg_3(i198):
        # Child args for sum_arg_3
        return((B_theta_coef_cp[i198]*n+(2-i198)*B_theta_coef_cp[i198])*X_coef_cp[n-i198+2])

    def sum_arg_2(i151):
        # Child args for sum_arg_2
        return(is_seq(0,n-i151)*diff(X_coef_cp[2*i151-n],True,1)*iota_coef[n-i151]*is_integer(n-i151)*is_seq(n-i151,i151))

    def sum_arg_1(i150):
        # Child args for sum_arg_1
        return((B_theta_coef_cp[i150]*n+(2-i150)*B_theta_coef_cp[i150])*Y_coef_cp[n-i150+2])


    out = -((is_seq(0,n+2)*Y_coef_cp[1]*dl_p*is_integer(n+2)*py_sum_parallel(sum_arg_62,0,n+2)-2*is_seq(0,n)*Y_coef_cp[1]*dl_p*is_integer(n)*py_sum_parallel(sum_arg_60,0,n)+is_seq(0,n+2)*X_coef_cp[1]*dl_p*is_integer(n+2)*py_sum_parallel(sum_arg_58,0,n+2)-2*is_seq(0,n)*X_coef_cp[1]*dl_p*is_integer(n)*py_sum_parallel(sum_arg_56,0,n)+(2*is_seq(0,n)*X_coef_cp[1]*dl_p*Y_coef_cp[n]-2*is_seq(0,n)*Y_coef_cp[1]*dl_p*X_coef_cp[n])*is_integer(n))*tau_p-X_coef_cp[1]*py_sum_parallel(sum_arg_7,ceil(n/2)+1,floor(n)+2)+2*is_seq(0,n)*Y_coef_cp[1]*is_integer(n)*py_sum_parallel(sum_arg_54,0,n)-2*is_seq(0,n)*Y_coef_cp[1]*is_integer(n)*py_sum_parallel(sum_arg_52,0,n)+2*is_seq(0,n)*X_coef_cp[1]*is_integer(n)*py_sum_parallel(sum_arg_50,0,n)-2*is_seq(0,n)*X_coef_cp[1]*is_integer(n)*py_sum_parallel(sum_arg_48,0,n)+Y_coef_cp[1]*py_sum_parallel(sum_arg_46,ceil(n/2)+1,floor(n)+2)+2*is_seq(0,n)*Y_coef_cp[1]*dl_p*is_integer(n)*py_sum_parallel(sum_arg_44,0,n)-is_seq(0,n+2)*Y_coef_cp[1]*is_integer(n+2)*py_sum_parallel(sum_arg_43,0,n+2)-Y_coef_cp[1]*py_sum_parallel(sum_arg_41,ceil(n/2)+1,floor(n)+2)+2*Y_coef_cp[1]*py_sum_parallel(sum_arg_4,ceil(n/2),floor(n))+is_seq(0,n+2)*X_coef_cp[1]*is_integer(n+2)*py_sum_parallel(sum_arg_39,0,n+2)+X_coef_cp[1]*py_sum_parallel(sum_arg_37,ceil(n/2)+1,floor(n)+2)-2*is_seq(0,n)*X_coef_cp[1]*dl_p*is_integer(n)*py_sum_parallel(sum_arg_35,0,n)-X_coef_cp[1]*py_sum_parallel(sum_arg_34,ceil(n/2)+1,floor(n)+2)+is_seq(0,n+2)*Y_coef_cp[1]*is_integer(n+2)*py_sum_parallel(sum_arg_32,0,n+2)+is_seq(0,n+2)*Y_coef_cp[1]*dl_p*kap_p*is_integer(n+2)*py_sum_parallel(sum_arg_30,0,n+2)-is_seq(0,n+2)*Y_coef_cp[1]*dl_p*is_integer(n+2)*py_sum_parallel(sum_arg_3,0,n+2)+is_seq(0,n+2)*Y_coef_cp[1]*dl_p*kap_p*is_integer(n+2)*py_sum_parallel(sum_arg_28,0,n+2)-2*is_seq(0,n)*Y_coef_cp[1]*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_26,0,n)-2*is_seq(0,n)*Y_coef_cp[1]*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_24,0,n)-is_seq(0,n+2)*X_coef_cp[1]*is_integer(n+2)*py_sum_parallel(sum_arg_22,0,n+2)-is_seq(0,n+2)*X_coef_cp[1]*dl_p*kap_p*is_integer(n+2)*py_sum_parallel(sum_arg_20,0,n+2)+2*X_coef_cp[1]*py_sum_parallel(sum_arg_2,ceil(n/2),floor(n))+2*is_seq(0,n)*X_coef_cp[1]*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_18,0,n)+Y_coef_cp[1]*py_sum_parallel(sum_arg_16,ceil(n/2)+1,floor(n)+2)-Y_coef_cp[1]*py_sum_parallel(sum_arg_13,ceil(n/2)+1,floor(n)+2)+X_coef_cp[1]*py_sum_parallel(sum_arg_10,ceil(n/2)+1,floor(n)+2)+is_seq(0,n+2)*X_coef_cp[1]*dl_p*is_integer(n+2)*py_sum_parallel(sum_arg_1,0,n+2)+2*is_seq(0,n)*Y_coef_cp[1]*is_integer(n)*diff(Y_coef_cp[n],False,1)+2*is_seq(0,n)*X_coef_cp[1]*is_integer(n)*diff(X_coef_cp[n],False,1)+2*is_seq(0,n)*X_coef_cp[1]*dl_p*kap_p*Z_coef_cp[n]*is_integer(n))/(2*dl_p*n+2*dl_p)
    return(out)

# Evaluates Zn. See Zn+1 for requirements.
def eval_Zn_cp(n, X_coef_cp, Y_coef_cp, Z_coef_cp,
    B_theta_coef_cp, B_psi_coef_cp,
    B_alpha_coef,
    kap_p, dl_p, tau_p, iota_coef):

    return(eval_Znp1_cp(n-1, X_coef_cp, Y_coef_cp, Z_coef_cp,
        B_theta_coef_cp, B_psi_coef_cp,
        B_alpha_coef,
        kap_p, dl_p, tau_p, iota_coef))



# D3_RHS[n] - D3_LHS[n] for calculating Y[n,+1].
# The constant component is what we need.
# Must evaluate with Yn, Zn+1=0
from math import floor, ceil
from aqsc.math_utilities import *
def eval_D3_RHS_m_LHS(n, X_coef_cp, Y_coef_cp, Z_coef_cp, B_theta_coef_cp,
    B_denom_coef_c, B_alpha_coef, iota_coef, dl_p, tau_p, kap_p):
    def sum_arg_15(i268):
        # Child args for sum_arg_15
        return(X_coef_cp[i268]*diff(Y_coef_cp[n-i268],True,1))

    def sum_arg_14(i264):
        # Child args for sum_arg_14
        return(Y_coef_cp[i264]*diff(X_coef_cp[n-i264],True,1))

    def sum_arg_13(i270):
        # Child args for sum_arg_13
        return(X_coef_cp[i270]*diff(Z_coef_cp[n-i270],True,1))

    def sum_arg_12(i266):
        # Child args for sum_arg_12
        return(Z_coef_cp[i266]*diff(X_coef_cp[n-i266],True,1))

    def sum_arg_11(i261):
        # Child args for sum_arg_11
        def sum_arg_10(i262):
            # Child args for sum_arg_10
            return(diff(Z_coef_cp[i262],True,1)*diff(Z_coef_cp[(-n)-i262+2*i261],True,1)*is_seq(n-i261,i261-i262))

        return(is_seq(0,n-i261)*iota_coef[n-i261]*is_integer(n-i261)*py_sum(sum_arg_10,0,i261))

    def sum_arg_9(i260):
        # Child args for sum_arg_9
        return(diff(Z_coef_cp[i260],True,1)*diff(Z_coef_cp[n-i260],False,1))

    def sum_arg_8(i257):
        # Child args for sum_arg_8
        def sum_arg_7(i258):
            # Child args for sum_arg_7
            return(diff(Y_coef_cp[i258],True,1)*diff(Y_coef_cp[(-n)-i258+2*i257],True,1)*is_seq(n-i257,i257-i258))

        return(is_seq(0,n-i257)*iota_coef[n-i257]*is_integer(n-i257)*py_sum(sum_arg_7,0,i257))

    def sum_arg_6(i256):
        # Child args for sum_arg_6
        return(diff(Y_coef_cp[i256],True,1)*diff(Y_coef_cp[n-i256],False,1))

    def sum_arg_5(i253):
        # Child args for sum_arg_5
        def sum_arg_4(i254):
            # Child args for sum_arg_4
            return(diff(X_coef_cp[i254],True,1)*diff(X_coef_cp[(-n)-i254+2*i253],True,1)*is_seq(n-i253,i253-i254))

        return(is_seq(0,n-i253)*iota_coef[n-i253]*is_integer(n-i253)*py_sum(sum_arg_4,0,i253))

    def sum_arg_3(i252):
        # Child args for sum_arg_3
        return(diff(X_coef_cp[i252],True,1)*diff(X_coef_cp[n-i252],False,1))

    def sum_arg_2(i209):
        # Child args for sum_arg_2
        def sum_arg_1(i210):
            # Child args for sum_arg_1
            return(B_theta_coef_cp[i210]*B_denom_coef_c[(-n)-i210+2*i209]*is_seq(n-i209,i209-i210))

        return(is_seq(0,n-i209)*B_alpha_coef[n-i209]*is_integer(n-i209)*py_sum(sum_arg_1,0,i209))


    out = ((is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_14,0,n)-is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_15,0,n))*tau_p)\
        +(is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_9,0,n))\
        +(py_sum_parallel(sum_arg_8,ceil(n/2),floor(n)))\
        +(is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_6,0,n))\
        +(py_sum_parallel(sum_arg_5,ceil(n/2),floor(n)))\
        +(is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_3,0,n))\
        +(-py_sum_parallel(sum_arg_2,ceil(n/2),floor(n)))\
        +(-is_seq(0,n)*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_13,0,n))\
        +(is_seq(0,n)*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_12,0,n))\
        +(py_sum_parallel(sum_arg_11,ceil(n/2),floor(n)))\
        +(is_seq(0,n)*dl_p*is_integer(n)*diff(Z_coef_cp[n],True,1))
    return(out)



import unittest
from aqsc import *
import pathlib

import jax.numpy as jnp

circ_equilibrium = circular_axis()
circ_axis_tolerance = 5e-6
def is_roughly_close_chiphifunc(chiphifunc_a, chiphifunc_b):
    return(
        jnp.all(
            jnp.abs(chiphifunc_a.content-chiphifunc_b.content)<circ_axis_tolerance
        )
    )

test_dir = str(pathlib.Path(__file__).parent.resolve())
B_psi_coef_cp, B_theta_coef_cp, \
    Delta_coef_cp, p_perp_coef_cp,\
    X_coef_cp, Y_coef_cp, Z_coef_cp, \
    iota_coef, dl_p,\
    nfp, Xi_0, eta, \
    B_denom_coef_c, B_alpha_coef, \
    kap_p, tau_p = read_first_three_orders(
        test_dir+'/circ/', 
        R_array=[2,0,1,2,0.0001,0],
        Z_array=[1,2,0,0.001]
    )

class TestCircularAxis(unittest.TestCase):

    def test_helicity(self):
        self.assertTrue(circ_equilibrium.get_helicity()==0)

    def test_B_psi(self):
        print('Testing leading order: B_psi[0]')
        print_fractional_error(
            circ_equilibrium.unknown['B_psi_coef_cp'][0].content,
            B_psi_coef_cp[0].content
        )
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['B_psi_coef_cp'][0],
            B_psi_coef_cp[0]
        ))

    def test_X(self):
        print('Testing leading order: X[1]')
        print_fractional_error(
            circ_equilibrium.unknown['X_coef_cp'][1].content,
            X_coef_cp[1].content
        )
        print('Testing leading order: X[2]')
        print_fractional_error(
            circ_equilibrium.unknown['X_coef_cp'][2].content,
            X_coef_cp[2].content
        )
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['X_coef_cp'][1],
            X_coef_cp[1]
        ))
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['X_coef_cp'][2],
            X_coef_cp[2]
        ))

    def test_Z(self):
        print('Testing leading order: Z[2]')
        print_fractional_error(
            circ_equilibrium.unknown['Z_coef_cp'][2].content,
            Z_coef_cp[2].content
        )
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['Z_coef_cp'][2],
            Z_coef_cp[2]
        ))

    def test_Y(self):
        print('Testing leading order: Y[1]')
        print_fractional_error(
            circ_equilibrium.unknown['Y_coef_cp'][1].content,
            Y_coef_cp[1].content
        )
        print('Testing leading order: Y[2]')
        print_fractional_error(
            circ_equilibrium.unknown['Y_coef_cp'][2].content,
            Y_coef_cp[2].content
        )
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['Y_coef_cp'][1],
            Y_coef_cp[1]
        ))
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['Y_coef_cp'][2],
            Y_coef_cp[2]
        ))

    def test_p(self):
        print('Testing leading order: p_perp[1]')
        print_fractional_error(
            circ_equilibrium.unknown['p_perp_coef_cp'][1].content,
            p_perp_coef_cp[1].content
        )
        print('Testing leading order: p_perp[2]')
        print_fractional_error(
            circ_equilibrium.unknown['p_perp_coef_cp'][2].content,
            p_perp_coef_cp[2].content
        )
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['p_perp_coef_cp'][1],
            p_perp_coef_cp[1]
        ))
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['p_perp_coef_cp'][2],
            p_perp_coef_cp[2]
        ))

    def test_Delta(self):
        print('Testing leading order: Delta[1]')
        print_fractional_error(
            circ_equilibrium.unknown['Delta_coef_cp'][1].content,
            Delta_coef_cp[1].content
        )
        print('Testing leading order: Delta[2]')
        print_fractional_error(
            circ_equilibrium.unknown['Delta_coef_cp'][2].content,
            Delta_coef_cp[2].content
        )
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['Delta_coef_cp'][1],
            Delta_coef_cp[1]
        ))
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['Delta_coef_cp'][2],
            Delta_coef_cp[2]
        ))

    
    def test_B_theta(self):
        print('Testing leading order: B_theta_coef_cp[2]')
        print_fractional_error(
            circ_equilibrium.unknown['B_theta_coef_cp'][2].content,
            B_theta_coef_cp[2].content
        )
        self.assertTrue(
            circ_equilibrium.unknown['B_theta_coef_cp'][1].is_special()
            and circ_equilibrium.unknown['B_theta_coef_cp'][1].nfp==0
        )
        self.assertTrue(is_roughly_close_chiphifunc(
            circ_equilibrium.unknown['B_theta_coef_cp'][2],
            B_theta_coef_cp[2]
        ))
        
if __name__ == '__main__':
    unittest.main()


''' Parsed files '''
# Coeff of B_psi terms and B_theta terms they carry in the looped equation.
# from 'Looped study B_theta through B_psi.wxmx'
from .lambda_B_psi_coef import *
# Coeff of B_psi in other variables for expediated evaluation. Also used in .
# from 'Looped study B_theta through B_psi.wxmx'
from .lambda_B_psi_coef_in_vars import *
# The coefficient of B_psi in D3.
from .B_psi_coefs_in_D3 import *
''' Hardcoded files '''
# The coefficient of Y1p in D3. Used in both MHD and magnetic iterations.
from .Y_coefs_in_D3 import *
# The coefficient of Y1p in the looped equation. Used in both MHD and magnetic iterations.
from .Y_coefs import *
# Coefficients of terms directly containing B_theta in the looped equation.
from .direct_B_theta_coefs import *



# RHS - LHS
# Used in (conv(a) + conv(b)@dchi)@Yn+1 = RHS - LHS(Yn+1 = 0)
# Must run with Yn+1=0.# Depends on Xn+1, Yn, Zn, B_theta n, B_psi n-2
# iota (n-2)/2 or (n-3)/2, B_alpha n/2 or (n-1)/2.
from math import floor, ceil
from aqsc.math_utilities import *
def rhs_minus_lhs(n,
    X_coef_cp,
    Y_coef_cp,
    Z_coef_cp,
    B_psi_coef_cp,
    B_theta_coef_cp,
    B_alpha_coef,
    kap_p, dl_p, tau_p,
    iota_coef):
    def sum_arg_31(i82):
        # Child args for sum_arg_31
        def sum_arg_30(i80):
            # Child args for sum_arg_30
            return(B_theta_coef_cp[i80]*X_coef_cp[n-i82-i80+2])

        return(i82*X_coef_cp[i82]*py_sum(sum_arg_30,0,n-i82+2))

    def sum_arg_29(i78):
        # Child args for sum_arg_29
        def sum_arg_28(i76):
            # Child args for sum_arg_28
            return(B_theta_coef_cp[i76]*Y_coef_cp[n-i78-i76+2])

        return(i78*Y_coef_cp[i78]*py_sum(sum_arg_28,0,n-i78+2))

    def sum_arg_27(i74):
        # Child args for sum_arg_27
        def sum_arg_26(i72):
            # Child args for sum_arg_26
            return(B_psi_coef_cp[i72]*X_coef_cp[n-i74-i72])

        return(diff(X_coef_cp[i74],True,1)*py_sum(sum_arg_26,0,n-i74))

    def sum_arg_25(i70):
        # Child args for sum_arg_25
        def sum_arg_24(i68):
            # Child args for sum_arg_24
            return(B_psi_coef_cp[i68]*Y_coef_cp[n-i70-i68])

        return(diff(Y_coef_cp[i70],True,1)*py_sum(sum_arg_24,0,n-i70))

    def sum_arg_23(i256):
        # Child args for sum_arg_23
        def sum_arg_22(i96):
            # Child args for sum_arg_22
            return(B_psi_coef_cp[i96]*diff(X_coef_cp[n-i96-i256],True,1))

        return(diff(Y_coef_cp[i256],False,1)*py_sum(sum_arg_22,0,n-i256))

    def sum_arg_21(i260):
        # Child args for sum_arg_21
        def sum_arg_20(i100):
            # Child args for sum_arg_20
            return(B_psi_coef_cp[i100]*diff(X_coef_cp[n-i260-i100],False,1))

        return(diff(Y_coef_cp[i260],True,1)*py_sum(sum_arg_20,0,n-i260))

    def sum_arg_19(i252):
        # Child args for sum_arg_19
        def sum_arg_18(i98):
            # Child args for sum_arg_18
            return((B_theta_coef_cp[i98]*n+((-i98)-i252+2)*B_theta_coef_cp[i98])*X_coef_cp[n-i98-i252+2])

        return(diff(Y_coef_cp[i252],False,1)*py_sum(sum_arg_18,0,n-i252+2))

    def sum_arg_17(i90):
        # Child args for sum_arg_17
        def sum_arg_16(i88):
            # Child args for sum_arg_16
            return((B_theta_coef_cp[i88]*n-B_theta_coef_cp[i88]*i90+(2-i88)*B_theta_coef_cp[i88])*Y_coef_cp[n-i90-i88+2])

        return(Z_coef_cp[i90]*py_sum(sum_arg_16,0,n-i90+2))

    def sum_arg_15(i86):
        # Child args for sum_arg_15
        def sum_arg_14(i84):
            # Child args for sum_arg_14
            return(B_psi_coef_cp[i84]*Z_coef_cp[n-i86-i84])

        return(diff(Y_coef_cp[i86],True,1)*py_sum(sum_arg_14,0,n-i86))

    def sum_arg_13(i257):
        # Child args for sum_arg_13
        def sum_arg_12(i258):
            # Child args for sum_arg_12
            return((((-i258)+2*i257-2)*diff(Y_coef_cp[i258],True,1)*X_coef_cp[(-n)-i258+2*i257-2]-diff(Y_coef_cp[i258],True,1)*X_coef_cp[(-n)-i258+2*i257-2]*n)*is_seq(n-i257+2,i257-i258))

        return(is_seq(0,n-i257+2)*B_alpha_coef[n-i257+2]*is_integer(n-i257+2)*py_sum(sum_arg_12,0,i257))

    def sum_arg_11(i109):
        # Child args for sum_arg_11
        def sum_arg_10(i110):
            # Child args for sum_arg_10
            return((((-i110)+2*i109-2)*diff(X_coef_cp[i110],True,1)*Y_coef_cp[(-n)-i110+2*i109-2]-diff(X_coef_cp[i110],True,1)*Y_coef_cp[(-n)-i110+2*i109-2]*n)*is_seq(n-i109+2,i109-i110))

        return(is_seq(0,n-i109+2)*B_alpha_coef[n-i109+2]*is_integer(n-i109+2)*py_sum(sum_arg_10,0,i109))

    def sum_arg_9(i106):
        # Child args for sum_arg_9
        def sum_arg_8(i104):
            # Child args for sum_arg_8
            return((B_theta_coef_cp[i104]*n-B_theta_coef_cp[i104]*i106+(2-i104)*B_theta_coef_cp[i104])*Y_coef_cp[n-i106-i104+2])

        return(diff(X_coef_cp[i106],False,1)*py_sum(sum_arg_8,0,n-i106+2))

    def sum_arg_7(i261):
        # Child args for sum_arg_7
        def sum_arg_6(i262):
            # Child args for sum_arg_6
            def sum_arg_5(i718):
                # Child args for sum_arg_5
                return(diff(X_coef_cp[(-i718)-i262+i261],True,1)*i718*Y_coef_cp[i718])

            return(is_seq(0,(-n)+i262+i261-2)*B_theta_coef_cp[(-n)+i262+i261-2]*is_integer((-n)+i262+i261-2)*is_seq((-n)+i262+i261-2,i262)*py_sum(sum_arg_5,0,i261-i262))

        return(iota_coef[n-i261+2]*py_sum(sum_arg_6,0,i261))

    def sum_arg_4(i253):
        # Child args for sum_arg_4
        def sum_arg_3(i254):
            # Child args for sum_arg_3
            def sum_arg_2(i702):
                # Child args for sum_arg_2
                return(diff(Y_coef_cp[(-i702)-i254+i253],True,1)*i702*X_coef_cp[i702])

            return(is_seq(0,(-n)+i254+i253-2)*B_theta_coef_cp[(-n)+i254+i253-2]*is_integer((-n)+i254+i253-2)*is_seq((-n)+i254+i253-2,i254)*py_sum(sum_arg_2,0,i253-i254))

        return(iota_coef[n-i253+2]*py_sum(sum_arg_3,0,i253))

    def sum_arg_1(i111):
        # Child args for sum_arg_1
        return(is_seq(0,n-i111)*diff(Z_coef_cp[2*i111-n],True,1)*iota_coef[n-i111]*is_integer(n-i111)*is_seq(n-i111,i111))


    out = -((is_seq(0,n+2)*dl_p*is_integer(n+2)*py_sum_parallel(sum_arg_31,0,n+2)+is_seq(0,n+2)*dl_p*is_integer(n+2)*py_sum_parallel(sum_arg_29,0,n+2)-2*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_27,0,n)-2*is_seq(0,n)*dl_p*is_integer(n)*py_sum_parallel(sum_arg_25,0,n))*tau_p+is_seq(0,n+2)*is_integer(n+2)*py_sum_parallel(sum_arg_9,0,n+2)+py_sum_parallel(sum_arg_7,ceil(n/2)+1,floor(n)+2)-py_sum_parallel(sum_arg_4,ceil(n/2)+1,floor(n)+2)+2*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_23,0,n)-2*is_seq(0,n)*is_integer(n)*py_sum_parallel(sum_arg_21,0,n)-is_seq(0,n+2)*is_integer(n+2)*py_sum_parallel(sum_arg_19,0,n+2)+is_seq(0,n+2)*dl_p*kap_p*is_integer(n+2)*py_sum_parallel(sum_arg_17,0,n+2)-2*is_seq(0,n)*dl_p*kap_p*is_integer(n)*py_sum_parallel(sum_arg_15,0,n)+py_sum_parallel(sum_arg_13,ceil(n/2)+1,floor(n)+2)-py_sum_parallel(sum_arg_11,ceil(n/2)+1,floor(n)+2)-2*py_sum_parallel(sum_arg_1,ceil(n/2),floor(n))-2*is_seq(0,n)*is_integer(n)*diff(Z_coef_cp[n],False,1)+2*is_seq(0,n)*dl_p*kap_p*X_coef_cp[n]*is_integer(n))/2
    return(out)


# Coefficient for Yn+1 - only X1 and Balpha0 needed
def coef_a(n, B_alpha_coef, X_coef_cp):

    out = (B_alpha_coef[0]*diff(X_coef_cp[1],True,1)*((-n)-1))/2
    return(out)


# Coefficient for dchi Yn+1  - only X1 and Balpha0 needed
def coef_b(B_alpha_coef, X_coef_cp):

    out = (B_alpha_coef[0]*X_coef_cp[1])/2
    return(out)
